# coding=utf-8
import re

import django
from django.apps import apps
from django.db.models import Model
from django.conf import settings as django_settings
from django.conf.urls import url
from django.contrib.admin import AdminSite, ModelAdmin
from django.contrib.admin.sites import site as django_admin_site
from django.contrib.auth.decorators import (
    user_passes_test, permission_required
)
from django.shortcuts import render
from django.utils.encoding import force_text
from django.utils.functional import cached_property
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _
from django.views.generic import View

from .environment import env
from . import compat


class ExtendedAdminSite(AdminSite):

    default_settings = {
        'INHERIT_REGISTERED_MODELS': env.get_bool(
            'DJANGO_ADMIN_INHERIT_REGISTERED_MODELS', True),
        'SITE_TITLE': env.get(
            'DJANGO_ADMIN_SITE_TITLE', AdminSite.site_title),
        'SITE_HEADER': env.get(
            'DJANGO_ADMIN_SITE_HEADER', 'admin'),
        'SITE_VERSION': env.get(
            'DJANGO_ADMIN_SITE_VERSION', django.__version__),
        'INDEX_TITLE': env.get(
            'DJANGO_ADMIN_INDEX_TITLE', AdminSite.index_title),
        'PRIMARY_COLOR': env.get(
            'DJANGO_ADMIN_PRIMARY_COLOR', '#34A77B'),
        'SECONDARY_COLOR': env.get(
            'DJANGO_ADMIN_SECONDARY_COLOR', '#20AA76'),
        'LOGO': env.get(
            'DJANGO_ADMIN_LOGO', 'admin/bananas/img/django.svg'),
        'LOGO_ALIGN': env.get(
            'DJANGO_ADMIN_LOGO_ALIGN', 'middle'),
        'LOGO_STYLE': env.get(
            'DJANGO_ADMIN_LOGO_STYLE'),
    }

    def __init__(self, name='admin'):
        super().__init__(name=name)
        self.settings = dict(self.default_settings)
        self.settings.update(getattr(django_settings, 'ADMIN', {}))

        self.site_title = self.settings['SITE_TITLE']
        self.site_header = self.settings['SITE_HEADER']
        self.index_title = self.settings['INDEX_TITLE']

    def each_context(self, request):
        context = super().each_context(request)
        context.update(settings=self.settings)
        return context

    @property
    def urls(self):
        if self.settings['INHERIT_REGISTERED_MODELS']:
            for model, admin in list(django_admin_site._registry.items()):
                # django_admin_site.unregister(model)
                self._registry[model] = admin.__class__(model, self)
        return self.get_urls(), 'admin', self.name


class ModelAdminView(ModelAdmin):

    @cached_property
    def access_permission(self):
        meta = self.model._meta
        return '{app_label}.{codename}'.format(
            app_label=meta.app_label,
            codename=meta.permissions[0][0],  # First perm codename
        )

    def get_urls(self):
        app_label = self.model._meta.app_label
        View = self.model.View
        info = app_label, View.label
        urlpatterns = compat.urlpatterns(
            url(r'^$',
                self.admin_view(View.as_view(admin=self)),
                name='{}_{}'.format(*info)),
            # We add the same url here with _changelist to make sure the
            # admin app index reverse urls to correct view.
            url(r'^$',
                self.admin_view(View.as_view(admin=self)),
                name='{}_{}_changelist'.format(*info)),
        )
        extra_urls = self.model.View(admin=self).get_urls()
        if extra_urls:
            urlpatterns += extra_urls
        return urlpatterns

    def admin_view(self, view, perm=None):
        if perm is not None:
            perm = self.get_permission(perm)
        else:
            perm = self.access_permission

        admin_login_url = compat.reverse_lazy('admin:login')
        view = user_passes_test(
            lambda u: u.is_active and u.is_staff,
            login_url=admin_login_url
        )(view)
        view = permission_required(perm, login_url=admin_login_url)(view)
        return view

    def get_permission(self, perm):
        if '.' not in perm:
            perm = '{}.{}'.format(self.model._meta.app_label, perm)
        return perm

    def has_module_permission(self, request):
        return request.user.has_perm(self.access_permission)

    def has_change_permission(self, request, obj=None):
        return request.user.has_perm(self.access_permission)

    def has_add_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False

    def get_context(self, request, **extra):
        opts = self.model._meta
        context = self.admin_site.each_context(request)
        context.update({
            'app_label': opts.app_label,
            'model_name': force_text(opts.verbose_name_plural),
            'title': force_text(opts.verbose_name_plural),
            'cl': {'opts': opts},  # change_list.html requirement
            'media': self.media,
        })
        context.update(extra or {})
        return context


def register(view=None, *, admin_site=None, admin_class=ModelAdminView):
    """
    Register a generic class based view wrapped with ModelAdmin and fake model

    :param view: The AdminView to register.
    :param admin_site: The AdminSite to register the view on.
        Defaults to bananas.admin.ExtendedAdminSite.
    :param admin_class: The ModelAdmin class to use for eg. permissions.
        Defaults to bananas.admin.ModelAdminView.

    Example:

    @register  # Or with args @register(admin_class=MyModelAdminSubclass)
    class MyAdminView(bananas.admin.AdminView):
        def get(request):
            return self.render('template.html', {})

    # Also possible:
    register(MyAdminView, admin_class=MyModelAdminSublass)

    """
    if not admin_site:
        admin_site = site

    def wrapped(inner_view):
        module = inner_view.__module__
        app_package = module[:module.index('.admin')]
        app_config = apps.get_app_config(app_package)

        label = getattr(inner_view, 'label', None)
        if not label:
            label = re.sub('(Admin)|(View)', '', inner_view.__name__).lower()
        inner_view.label = label

        model_name = label.capitalize()
        verbose_name = getattr(inner_view, 'verbose_name', model_name)
        inner_view.verbose_name = verbose_name

        access_perm_codename = 'can_access_' + model_name.lower()
        access_perm_name = _('Can access {verbose_name}').format(
            verbose_name=verbose_name
        )
        # The first permission here is expected to be
        # the general access permission.
        permissions = tuple(
            [(access_perm_codename, access_perm_name)] +
            list(getattr(inner_view, 'permissions', []))
        )

        model = type(model_name, (Model,), {
            '__module__': module + '.__models__',  # Fake
            'View': inner_view,
            'app_config': app_config,
            'Meta': type('Meta', (object,), dict(
                managed=False,
                abstract=True,
                app_label=app_config.label,
                verbose_name=verbose_name,
                verbose_name_plural=verbose_name,
                permissions=permissions,
            )),
        })

        admin_site._registry[model] = admin_class(model, admin_site)
        return inner_view

    if view is None:  # Used as a decorator
        return wrapped

    return wrapped(view)


class ViewTool:
    def __init__(self, text, link, perm=None, **attrs):
        self.text = text
        self._link = link
        self.perm = perm

        html_class = attrs.pop('html_class', None)
        if html_class is not None:
            attrs.setdefault('class', html_class)
        self._attrs = attrs

    @property
    def attrs(self):
        return mark_safe(
            ' '.join('{}={}'.format(k, v) for k, v in self._attrs.items())
        )

    @property
    def link(self):
        if '/' not in self._link:
            return compat.reverse(self._link)
        return self._link


class AdminView(View):
    admin = None  # type: ModelAdminView
    tools = None
    action = None

    def dispatch(self, request, *args, **kwargs):
        # Try to fetch set action first.
        # This should be the view name for custom views
        action = self.action
        if action is None:
            return super().dispatch(request, *args, **kwargs)

        handler = getattr(self, action)
        return handler(request, *args, **kwargs)

    def get_urls(self):
        """ Should return a list of urls
        Views should be wrapped in `self.admin_view` if the view isn't
        supposed to be accessible for non admin users.
        Omitting this can cause threading issues.

        Example:

            return [
                url(r'^custom/$,
                    self.admin_view(self.custom_view))
            ]
        """
        return None

    def get_tools(self):
        # Override point, self.request is available.
        return self.tools

    def get_view_tools(self):
        tools = []
        all_tools = self.get_tools()
        if all_tools:
            for tool in all_tools:
                if isinstance(tool, (list, tuple)):
                    perm = None
                    if len(tool) == 3:
                        tool, perm = tool[:-1], tool[-1]
                    text, link = tool
                    tool = ViewTool(text, link, perm=perm)
                else:
                    # Assume ViewTool
                    perm = tool.perm
                if perm and not self.has_permission(perm):
                    continue

                tools.append(tool)

        return tools

    def admin_view(self, view, perm=None, **initkwargs):
        view = self.__class__.as_view(
            action=view.__name__,
            admin=self.admin,
            **initkwargs
        )
        return self.admin.admin_view(view, perm=perm)

    def get_permission(self, perm):
        return self.admin.get_permission(perm)

    def has_permission(self, perm):
        perm = self.get_permission(perm)
        return self.request.user.has_perm(perm)

    def has_access(self):
        return self.has_permission(self.admin.access_permission)

    def get_context(self, **extra):
        return self.admin.get_context(
            self.request,
            view_tools=self.get_view_tools(),
            **extra
        )

    def render(self, template, context=None):
        extra = context or {}
        return render(
            self.request,
            template,
            self.get_context(**extra)
        )


site = ExtendedAdminSite()



from django.contrib.auth.models import Permission
from django.core.management.base import BaseCommand, CommandError


class Command(BaseCommand):

    help = "Create admin permissions"

    def handle(self, *args, **options):
        if args:  # pragma: no cover
            raise CommandError("Command doesn't accept any arguments")
        return self.handle_noargs(**options)

    def handle_noargs(self, *args, **options):
        from bananas import admin
        from django.contrib import admin as django_admin
        from django.contrib.contenttypes.models import ContentType

        django_admin.autodiscover()

        for model, _ in admin.site._registry.items():
            if issubclass(getattr(model, 'View', object), admin.AdminView):
                meta = model._meta

                ct, created = ContentType.objects.get_or_create(
                    app_label=meta.app_label,
                    model=meta.object_name.lower(),
                )

                if created:
                    print('Found new admin view: {} [{}]'.format(
                        ct.name, ct.app_label
                    ))

                for codename, name in model._meta.permissions:
                    p, created = Permission.objects.get_or_create(
                        codename=codename, name=name, content_type=ct
                    )
                    if created:
                        print('Created permission: {}'.format(name))



import django

__all__ = (
    'get_resolver', 'reverse', 'reverse_lazy', 'URLPattern',
    'URLResolver', 'urlpatterns'
)

if django.VERSION < (2, 0):
    from django.core.urlresolvers import get_resolver, reverse, reverse_lazy
    from django.conf.urls import RegexURLPattern as URLPattern, \
        RegexURLResolver as URLResolver
else:
    from django.urls import (
        get_resolver, reverse, reverse_lazy, URLPattern, URLResolver
    )


def urlpatterns(*urls):
    if django.VERSION < (1, 10):
        from django.conf.urls import patterns
        return patterns('', *urls)

    return list(urls)



from django.db.models.query import ValuesQuerySet

from bananas.models import ModelDict


class ExtendedValuesQuerySet(ValuesQuerySet):
    """
    Extended `ValuesQuerySet` with support for renaming fields
    and choice of object class.
    """
    _values_class = dict

    @property
    def named_fields(self):
        return getattr(self, '_named_fields')

    def rename_fields(self, names):
        named_fields = {value: key for key, value in self.named_fields.items()}
        names = [named_fields.get(name, name) for name in names]
        return names

    def iterator(self):
        # Purge any extra columns that haven't been explicitly asked for
        extra_names = list(self.query.extra_select)
        field_names = self.field_names
        annotation_names = list(self.query.annotation_select)

        # Modified super(); rename fields given in queryset.values() kwargs
        names = self.rename_fields(extra_names + field_names + annotation_names)
        if self.named_fields:
            names = self.rename_fields(names)

        for row in self.query.get_compiler(self.db).results_iter():
            yield self._values_class(zip(names, row))

    def _clone(self, klass=None, setup=False, **kwargs):
        kwargs.update(_named_fields=self.named_fields,
                      _values_class=self._values_class,
                      klass=klass,
                      setup=setup)

        return super()._clone(**kwargs)


class ModelDictQuerySetMixin:

    def dicts(self, *fields, **named_fields):
        if named_fields:
            fields += tuple(named_fields.values())

        return self._clone(klass=ExtendedValuesQuerySet, setup=True,
                           _fields=fields, _named_fields=named_fields,
                           _values_class=ModelDict)



import sys
from collections import OrderedDict

import django
from django.core.management import BaseCommand
from ... import compat


def collect_urls(urls=None, namespace=None, prefix=None):
    if urls is None:
        urls = compat.get_resolver(urlconf=None)
    prefix = prefix or []
    if isinstance(urls, compat.URLResolver):
        res = []
        if django.VERSION < (2, 0):
            pattern = urls.regex.pattern
        else:
            pattern = urls.pattern.regex.pattern
        for x in urls.url_patterns:
            res += collect_urls(x, namespace=urls.namespace or namespace,
                                prefix=prefix + [pattern])
        return res
    elif isinstance(urls, compat.URLPattern):
        if django.VERSION < (1, 10):
            callback = urls._callback
            lookup_str = callback.__module__ + '.'
            if hasattr(callback, '__name__'):
                lookup_str += callback.__name__
            else:
                lookup_str += callback.__class__.__name__
        else:  # pragma: no cover
            lookup_str = urls.lookup_str

        if django.VERSION < (2, 0):
            pattern = urls.regex.pattern
        else:
            pattern = urls.pattern.regex.pattern
        return [OrderedDict([('namespace', namespace),
                             ('name', urls.name),
                             ('pattern', prefix + [pattern]),
                             ('lookup_str', lookup_str),
                             ('default_args', dict(urls.default_args))])]
    else:  # pragma: no cover
        raise NotImplementedError(repr(urls))


def show_urls():
    all_urls = collect_urls()

    max_lengths = {}
    for u in all_urls:
        for k in ['pattern', 'default_args']:
            u[k] = str(u[k])
        for k, v in list(u.items())[:-1]:
            u[k] = v = v or ''
            # Skip app_list due to length (contains all app names)
            if (u['namespace'], u['name'], k) == \
                    ('admin', 'app_list', 'pattern'):
                continue
            max_lengths[k] = max(len(v), max_lengths.get(k, 0))

    for u in sorted(all_urls, key=lambda x: (x['namespace'], x['name'])):
        sys.stdout.write(' | '.join(
            ('{:%d}' % max_lengths.get(k, len(v))).format(v)
            for k, v in u.items()) + '\n')


class Command(BaseCommand):
    def handle(self, *args, **kwargs):
        show_urls()



import logging

import django
from django.db.models.query import QuerySet

from .models import ModelDict


_log = logging.getLogger(__name__)


class ModelDictIterable(object):

    def __init__(self, queryset):
        self.queryset = queryset
        self.named_fields = self.queryset._hints.get('_named_fields')

    def __iter__(self):
        queryset = self.queryset
        query = queryset.query
        compiler = query.get_compiler(queryset.db)

        field_names = list(query.values_select)
        extra_names = list(query.extra_select)
        annotation_names = list(query.annotation_select)

        # Modified super(); rename fields given in queryset.values() kwargs
        names = extra_names + field_names + annotation_names
        if self.named_fields:
            names = self.rename_fields(names)

        for row in compiler.results_iter():
            yield ModelDict(zip(names, row))

    def rename_fields(self, names):
        named_fields = {value: key for key, value in self.named_fields.items()}
        names = [named_fields.get(name, name) for name in names]
        return names


class ModelDictQuerySetMixin:
    """
    Supported by Django >= 1.9
    """
    def dicts(self, *fields, **named_fields):
        if named_fields:
            fields += tuple(named_fields.values())

        clone = self.values(*fields)
        clone._iterable_class = ModelDictIterable

        # QuerySet._hints is a dict object used by db router
        # to aid deciding which db should get a request. Currently
        # django only supports `instance`, so it's probably
        # fine to set a custom key on this dict as it's a guaranteed
        # way that it'll be returned with the QuerySet instance
        # while leaving the queryset intact
        clone._add_hints(**{'_named_fields': named_fields})

        return clone


if django.VERSION < (1, 9):
    """
    Patch ModelDictQuerySetMixin for old Django compatibility
    """
    from .compat.django18 import ModelDictQuerySetMixin  # noqa


class ModelDictQuerySet(ModelDictQuerySetMixin, QuerySet):
    pass


class ModelDictManagerMixin:

    def dicts(self, *fields, **named_fields):
        return self.get_queryset().dicts(*fields, **named_fields)

    def get_queryset(self):
        return ModelDictQuerySet(self.model, using=self._db)


ExtendedQuerySet = ModelDictQuerySet  # Left for compatibility



import logging

from functools import partial
from os import environ

from django.conf import global_settings


__all__ = [
    'parse_bool', 'parse_int', 'parse_tuple', 'parse_list', 'parse_set'
]

log = logging.getLogger(__name__)

UNDEFINED = object()

UNSUPPORTED_ENV_SETTINGS = (
    'ADMINS',
    'MANAGERS',
    'LANGUAGES',
    'DISALLOWED_USER_AGENTS',
    'IGNORABLE_404_URLS',
    'TEMPLATES',
)

SETTINGS_TYPES = {
    'LANGUAGE_COOKIE_AGE': int,
    'EMAIL_TIMEOUT': int,
    'FILE_UPLOAD_PERMISSIONS': int,
    'FILE_UPLOAD_DIRECTORY_PERMISSIONS': int,
}


def parse_str(value):
    """
    Clean string.

    :param str value: Original str.
    :return: str: Cleaned str.
    """
    return value.strip()


def parse_bool(value):
    """
    Parse string to bool.

    :param str value: String value to parse as bool
    :return bool:
    """
    boolean = parse_str(value).capitalize()

    if boolean in ('True', 'Yes', 'On', '1'):
        return True
    elif boolean in ('False', 'No', 'Off', '0'):
        return False
    else:
        raise ValueError('Unable to parse boolean value "{}"'.format(value))


def parse_int(value):
    """
    Parse numeric string to int. Supports oct formatted string.

    :param str value: String value to parse as int
    :return int:
    """
    value = parse_str(value=value)
    if value.startswith('0'):
        return int(value.lstrip('0o'), 8)
    else:
        return int(value)


def parse_iterable(typ, value):
    """
    Parse comma separated string into typed iterable.

    :param typ: Type to cast iterable to
    :param value: String value to parse as iterable
    :return: Given type
    """
    return typ(parse_str(v) for v in value.split(','))


parse_tuple = partial(parse_iterable, tuple)
parse_list = partial(parse_iterable, list)
parse_set = partial(parse_iterable, set)


def get_parser(typ):
    """
    Return appropriate parser for given type.

    :param typ: Type to get parser for.
    :return function: Parser
    """
    try:
        return {
            str: parse_str,
            bool: parse_bool,
            int: parse_int,
            tuple: parse_tuple,
            list: parse_list,
            set: parse_set,
        }[typ]
    except KeyError:
        raise NotImplementedError("Unsupported setting type: %r", typ)


def get_settings():
    """
    Get and parse prefixed django settings from env.

    TODO: Implement support for complex settings
        DATABASES = {}
        CACHES = {}
        INSTALLED_APPS -> EXCLUDE_APPS ?

    :return dict:
    """
    settings = {}
    prefix = environ.get('DJANGO_SETTINGS_PREFIX', 'DJANGO_')

    for key, value in environ.items():
        _, _, key = key.partition(prefix)
        if key:
            if key in UNSUPPORTED_ENV_SETTINGS:
                raise ValueError('Django setting "{}" can not be '
                                 'configured through environment.'.format(key))

            default_value = getattr(global_settings, key, UNDEFINED)

            if default_value is not UNDEFINED:
                if default_value is None and key in SETTINGS_TYPES.keys():
                    # Handle typed django settings defaulting to None
                    parse = get_parser(SETTINGS_TYPES[key])
                else:
                    # Determine parser by django setting type
                    parse = get_parser(type(default_value))

                value = parse(value)

            settings[key] = value

    return settings


class EnvironWrapper(object):
    """
    Wrapper around os environ with type conversion support.
    """
    __delitem__ = environ.__delitem__
    __getitem__ = environ.__getitem__
    __setitem__ = environ.__setitem__
    __contains__ = environ.__contains__

    def __getattr__(self, item):
        return getattr(environ, item)

    def parse(self, parser, key, default=None):
        value = environ.get(key, UNDEFINED)
        if value is UNDEFINED:
            return default
        try:
            return parser(value)
        except ValueError:
            log.warning((
                'Unable to parse environment variable '
                '{key}={value}'
            ).format(
                key=key,
                value=value,
            ))
            return default

    def get_bool(self, key, default=None):
        return self.parse(parse_bool, key, default=default)

    def get_int(self, key, default=None):
        return self.parse(parse_int, key, default=default)

    def get_tuple(self, key, default=None):
        return self.parse(parse_tuple, key, default=default)

    def get_list(self, key, default=None):
        return self.parse(parse_list, key, default=default)

    def get_set(self, key, default=None):
        return self.parse(parse_set, key, default=default)


env = EnvironWrapper()
